/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>,
 * Ryan Barrett <p4@ryanb.org>,
 * Maulik Shah <p4@maulik.net>,
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file p4plugin.h
 *
 * The header that P4 plugins should use to compile. Exports all of the API
 * functions as C prototypes.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef _P4_PLUGIN_H
#define _P4_PLUGIN_H

#include <sys/types.h>

#ifndef P4BACKENDEXPORT
#ifdef _WIN32
#define P4BACKENDEXPORT extern "C" __declspec(dllexport)
#else
#define P4BACKENDEXPORT
#endif
#endif

/*****************************************************************************
 * CONSTANTS
 *****************************************************************************/
#define kp4_username_len 20  // includes null terminator!!!


/*****************************************************************************
 * PLUGIN API FUNCTIONS
 *****************************************************************************/

/** Many of the plugin API functions take a ushort plugin_id parameter. The
 * plugin should pass its own plugin id so that messages can be correctly
 * identified and routed.
 *
 * All of the plugin API functions return 0 on success, >0 on failure. The
 * error codes can be found in error.h. To print an error, use the p4_perror
 * function.
 */

/** p4_broadcast
 *
 * This call will ensure that data gets to every user on the network. The
 * length of the data in bytes is passed in size.
 */

P4BACKENDEXPORT uint p4_broadcast(ushort plugin_id, const char *data, int size,
                                  bool encrypted);


/** p4_send
 *
 * This call will transfer data to user through a direct connection. The length
 * of the data in bytes is passed in size. If encrypted is true, the data will
 * be encrypted with the recipient's public key.
 */

P4BACKENDEXPORT uint p4_send(ushort plugin_id, const char *user,
                             const char *data, int size, bool encrypted);


/** p4_get_num_plugins
 *
 * Writes the number of active plugins that the given user has into
 * num_plugins.
 */

P4BACKENDEXPORT uint p4_get_num_plugins(const char *user, ushort *num_plugins);


/** p4_get_plugins
 *
 * p4_get_plugins tells which plugins user is running. Plugin_ids must be an
 * array of ushorts since only id numbers are necessary for plugins.
 * num_written will contain the number of plugin names written to plugins.
 */

P4BACKENDEXPORT uint p4_get_plugins(const char *user, ushort *plugin_ids,
                                    ushort num_plugins, ushort *num_written);


/** p4_get_latency
 *
 * Reports the latency in milliseconds to the user. The latency value is
 * written into latency.
 */

P4BACKENDEXPORT uint p4_get_latency(const char *user, int *latency);


/** p4_get_num_users
 *
 * Writes to num_users the number of users running the given plugin. If the
 * plugin_id field is 0, the total number of users logged on is written.
 */

P4BACKENDEXPORT uint p4_get_num_users(ushort plugin_id, int *num_users);


/** p4_get_users
 *
 * Writes num_users worth of usernames to users who are running plugin_id. The
 * actual number written is placed in num_written. If the plugin_id field is
 * NULL, all users logged on are written to users.
 */

P4BACKENDEXPORT uint p4_get_users(ushort plugin_id, char **users,
                                  int num_users, int *num_written);


/** p4_authenticate
 *
 * Authenticates that the last message received by the calling plugin was
 * actually sent by the said user. If the signature is valid, true is written
 * to authenticated, false otherwise.
 */

P4BACKENDEXPORT uint p4_authenticate(ushort plugin_id, bool *authenticated);


/** p4_peek
 *
 * Writes to size the number of bytes of the next message. If no message is
 * waiting, 0 is written.
 */

P4BACKENDEXPORT uint p4_peek(ushort plugin_id, int *size);


/** p4_receive
 *
 * Writes to the array allocated for user the name of the user who sent the
 * message. Writes to data at most size bytes of the message. The actual amount
 * copied is written to size_written. encrypted tells whether the message
 * was encrypted when it arrived.
 */

P4BACKENDEXPORT uint p4_receive(ushort plugin_id, char *user, char *data,
                                int size, int *size_written, bool *encrypted);



#endif  // _P4_PLUGIN_H


