/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file p4client.cpp
 *
 * Implements the client API functions. For more information, see p4client.h.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "conn_mgr.h"
#include "msg_hello.h"
#include "msg_lka_ack.h"
#include "msg_lka.h"
#include "msg_connect.h"
#include "conn_mgr.h"
#include "lka.h"
#include "receiveq.h"
#include "p4error.h"
#include "p4client.h"
#include "servicer.h"
#include "sys.h"

#ifdef _WIN32  // for gettimeofday
#include <winsock.h>
#include <sys/timeb.h>
#else
#include <arpa/inet.h>
#include <sys/time.h>
#endif

#include <errno.h>
#include <string>

/**********************************************************************
 * GLOBALS
 **********************************************************************/
// these are all for bootstrap
extern string glocal_user;
extern pthread_t servicer;  // the main thread for the backend

static stop_info stop;
bool done_bootstrapping = false;

/**********************************************************************
 * HELPER PROTOTYPES
 **********************************************************************/
static uint get_lka(const string &username);
static uint connect(const string &username, const string &address, 
                    ushort port);
static bool get_remote_host_name(string &username, string &address, 
                                 ushort &port, int i);


/**********************************************************************
 * CLIENT API FUNCTIONS
 **********************************************************************/
uint p4_bootstrap()
{
  int num_connections = csettings::instance().get<int>("num_boot_connections");
  uint ret = 0;
  bool connected = false;

  // for the remote user
  string username, address;
  ushort port;

  // seed the random number generator!
  time_t curr;
  curr = time(&curr);
  srand(curr);  

#ifdef _WIN32
  WSADATA wsaData;    // if this doesn't work
  // WSAData wsaData; // then try this instead

  if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0) {
    cerr << "WSAStartup failed." << endl;
    return kerr_boot_init_failed;
  } 
#endif

  stop.time_to_stop = false;
  sem_init(&stop.sem, 0, 0);

  // enter local user into lka
  clka::instance().update_local_user();

  // start the connection sequence
  ret = cconn_mgr::instance().listen();
  if (ret == kerr_success) {
    cdebug::instance() << acquire("servicer_debug")
                       << "Listening for connections"
                       << release();

    // start the servicer loop (used to use an attribute)
    if (pthread_create(&servicer, NULL, &servicer_loop, 
                       reinterpret_cast<void*>(&stop))) {
      cerr << "Could not create servicer thread" << endl;
      return kerr_boot_init_failed;
    }
    
    // try to connect to each of the entries in .p4
    // stop after first one succeeds
    for (int i = 0; i < num_connections; i++) {
      if (!get_remote_host_name(username, address, port, i)) {
        cdebug::instance() << acquire("servicer_debug")
                           << "Could not read info for remote #" << i
                           << release();
      }

      ret = connect(username, address, port);
      if (ret == kerr_success) {
        cdebug::instance() << acquire("servicer_debug")
                           << "Connected to " << username
                           << release();
        connected = true;
        break;
      }
    }
    // don't delete servicer
    done_bootstrapping = true;
    return connected ? kerr_success : kerr_couldnt_connect;
  }

  cdebug::instance() << acquire("servicer_debug")
                     << "Could not set up listener"
                     << release();
  return kerr_boot_init_failed;
}

/** p4_add_connection
 * 
 * Makes sure that we aren't already connected to the network before trying the
 * new connection. If we were to connect to a new client while we have other
 * connections open, the new and old connections would not learn about each
 * other.
 */

uint p4_add_connection(const char *username, const char *addr, ushort port)
{
  cdebug::instance() << acquire("servicer_debug")
                     << "p4_add_connection() to" << username << ", " << addr
                     << ":" << port
                     << release();
  if (strlen(username) == 0 || strlen(addr) == 0 || port < 1024)
    return kerr_bad_arg;

  if (cconn_mgr::instance().num_connected() > 0)
    return kerr_already_connected;

  clka::instance().wipe();

  // enter local user into lka
  clka::instance().update_local_user();

  return connect(username, addr, port);
}


uint p4_quit()
{
  cdebug::instance() << acquire("servicer_debug")
                     << "p4_quit() called, stopping network service."
                     << release();

  stop.time_to_stop = true;
  sem_wait(&stop.sem);
  return kerr_success;
}


uint p4_set_username(const char *username)
{
  glocal_user = username;
  return kerr_success;
}


const char *p4_get_username()
{
  return glocal_user.c_str();
}


uint p4_plugin_loaded(ushort plugin_id)
{
  return creceiveq::instance().plugin_loaded(plugin_id);
}


uint p4_plugin_stopped(ushort plugin_id)
{
  return creceiveq::instance().plugin_stopped(plugin_id);
}


/*****************************************************************************
 * CONNECTING FUNCTIONS
 *****************************************************************************/

/** get_lka
 *
 * Sends a msg_lka and then waits on a condition in cmsg_lka_ack until we
 * get one.
 */
static uint get_lka(const string &username)
{
  cmsg_lka lka_request;
  const cconn &conn = lka_request.send(username);
  conn.wait_until_ready();
  if (conn.get_status() == cconn::efailed) {
    return kerr_couldnt_get_lka;
  }
  
  // set the timeout on the condition
#ifdef _WIN32
  struct timespec now;
#else
  struct timeval now;
#endif
  struct timespec timeout;
  int retcode;
  
  if (gettimeofday(&now, NULL))
    return kerr_couldnt_get_lka;
  timeout.tv_sec = now.tv_sec + 
    csettings::instance().get<uint>("lka_wait_time");
#ifdef _WIN32
  timeout.tv_nsec = now.tv_nsec * 1000;
#else
  timeout.tv_nsec = now.tv_usec * 1000;
#endif
  retcode = 0;

  // see if we already have an lka
  pthread_mutex_lock(&cmsg_lka_ack::received_mtx);    

  if (!cmsg_lka_ack::received) {  
    retcode = pthread_cond_timedwait(&cmsg_lka_ack::received_cond,
                                     &cmsg_lka_ack::received_mtx, &timeout);
    
    if (retcode == ETIMEDOUT || retcode == EINTR) {  // timeout occurred 
      cdebug::instance() << acquire("servicer_debug")
                         << "waiting for lka_ack timed out"
                         << release();
      pthread_mutex_unlock(&cmsg_lka_ack::received_mtx);
      return kerr_couldnt_get_lka;
    } else {  // lka_request succeeded
      cdebug::instance() << acquire("servicer_debug")
                         << "got an lka"
                         << release();
      assert(cmsg_lka_ack::received);
    }
  }

  pthread_mutex_unlock(&cmsg_lka_ack::received_mtx);
  return kerr_success;
}

/** connect
 *
 * Used to open the first connection for this client.  Resolves the ip
 * address, sends a connect, lka, and hello messages.
 */
static uint connect(const string &username, const string &address,
          ushort port)
{
  cdebug::instance() << acquire("servicer_debug")
                     << "trying to connect to " << address
                     << release();

  // resolve address
  struct in_addr ip = resolve_addr(address.c_str());
  if (ip.s_addr == INADDR_NONE) {
    cdebug::instance() << acquire("servicer_debug")
                       << "couldn't resolve addr"
                       << release();
    return kerr_bad_address;
  }


  // add to LKA
  if (clka::instance().add(username.c_str(), ip.s_addr, port, 0, 0)
      != kerr_success) {
    cdebug::instance() << acquire("servicer_debug")
                       << "Could not add " << username << " to lka"
                       << release();
    return kerr_couldnt_add_user;
  }


    // try to connect now
  cmsg_connect connect;
  const cconn &conn = connect.send(username);

  conn.wait_until_ready();
  if (conn.get_status() == cconn::efailed) {
    cdebug::instance() << acquire("servicer_debug")
                       << "Didn't connect to " << username
                       << release();
    return kerr_couldnt_connect;
  }

  cdebug::instance() << acquire("servicer_debug")
                     << "Connected to " << username
                     << release();

  // try to get an lka from another host
  uint ret = get_lka(username);
  if (ret != kerr_success) {
    cdebug::instance() << acquire("servicer_debug")
                       << "Could not get lka from " << address
                       << release();
    return ret;
  }

  // Send hello
  cmsg_hello hello;
  
  const cconn &conn2 = hello.send(username);
  conn2.wait_until_ready();
  if (conn2.get_status() == cconn::efailed) {
    cdebug::instance() << acquire("servicer_debug")
                       << "Could not say hello to " << address
                       << release();
    return kerr_couldnt_connect;
  }
  cdebug::instance() << acquire("servicer_debug")
                     << "Said hello to " << address
                     << release();

  return kerr_success;
}

/** get_remote_host_name
 *
 * looks up the ith host in the settings to figure out the hostname,
 * user, and port on that host. 
 */
bool get_remote_host_name(string &username, string &address,
              ushort &port, int i)
{
  char c = '0' + i;

  address = csettings::instance().get<string>(string("remote_host") + c);
  port = csettings::instance().get<ushort>(string("remote_port") + c);
  username = csettings::instance().get<string>(string("remote_user") + c);

  if (address == "" || port == 0 || username == "") {
    cdebug::instance() << acquire("servicer_debug")
                       << "Could not load remote host " << c
                       << release();
    return false;
  }
  
  return true;
}


